(defstruct word
  (word nil :type string))

(defun show-word (word)
  (if (word-p word)
	  (word-word word)
	  word))

(defstruct state
  (words nil :type list)
  (transitions nil :type hash-table)
  (no-transitions nil :type integer))

(defun new-state (words)
  (make-state :words words
			  :transitions (make-hash-table)
			  :no-transitions 0))

(defstruct chain
  (order 0 :type integer)
  start
  end
  (words nil :type hash-table)
  (states nil :type hash-table)
  (start-states nil :type hash-table)
  (no-start-states nil :type integer))

(defun new-chain (order)
  (make-chain :order order
			  :start 'start
			  :end 'end
			  :words (make-hash-table :test #'equal)
			  :states (make-hash-table :test #'equal)
			  :start-states (make-hash-table)
			  :no-start-states 0))

(defun print-chain (chain)
  (loop for s1 being the hash-values in (chain-states chain)
	   do (loop for s2 being the hash-keys in (state-transitions s1) using (hash-value v)
			   do (format t "~a -~a-> ~a~%" 
						  (mapcar #'show-word (state-words s1))
						  v
						  (mapcar #'show-word (state-words s2))))))

(defmacro hash-default-insert (hash-table key value)
  `(multiple-value-bind (v exists) (gethash ,key ,hash-table)
	 (if exists
		 v
		 (setf (gethash ,key ,hash-table) ,value))))

(defmacro hash-map-key (hash-table key f value)
  `(multiple-value-bind (v exists) (gethash ,key ,hash-table)
	 (if exists
		 (setf (gethash ,key ,hash-table) (,f v))
		 (setf (gethash ,key ,hash-table) ,value))))

(defun get-word (chain word)
  (hash-default-insert (chain-words chain) word (make-word :word word)))

(defun get-state (chain words)
  (hash-default-insert (chain-states chain) words (new-state words)))

(defun first-n (l n)
  (if (eq n 0)
	  nil
	  (cons (car l) (first-n (cdr l) (- n 1)))))

(defun train (chain s1 s2)
  (hash-map-key (state-transitions s1) s2 1+ 1)
  (incf (state-no-transitions s1))
  (if (eq (chain-start chain) (first (state-words s1)))
	  (progn (hash-map-key (chain-start-states chain) s1 1+ 1)
		 (incf (chain-no-start-states chain)))))

(defun train-list (chain all-words)
  (loop for words = all-words then (cdr words)
	 while (>= (length words) (1+ (chain-order chain)))
	 do (train chain
			   (get-state chain (first-n words 
										 (chain-order chain)))
			   (get-state chain (first-n (cdr words) 
										 (chain-order chain))))))

(defun train-sentence (chain words)
  (train-list chain (append (list (chain-start chain))
							(mapcar (lambda (word) (get-word chain word)) words)
							(list (chain-end chain)))))

(defun train-from-stdin (chain)
  (loop with sentence = '()
	 for line = (read-line nil nil)
	 while line
	 do (if (eq 0 (length line))
			(if sentence
				(progn (train-sentence chain (reverse sentence))
					   (setf sentence '())))
			(push line sentence))))
				

;; Generation

(defun random-pick (hash total)
  (loop for n = (random total) then (- n v)
	 for next-state being the hash-keys in hash using (hash-value v)
	 when (< n v) return next-state))


(defun generate (chain)
  (rest (loop for state = (random-pick (chain-start-states chain)
									   (chain-no-start-states chain))
		                  then (random-pick (state-transitions state)
											(state-no-transitions state))
		   collecting state
		   until (eq (car (last (state-words state)))
					 (chain-end chain)))))

(defun main ()
  (let ((chain (new-chain 2)))
	(train-from-stdin chain)
	(print (generate chain))))

(main)